import numpy as np

# Load MNIST data from local .npz file
def load_mnist(path='mnist.npz'):
    data = np.load(path)
    x_train, y_train = data['x_train'], data['y_train']
    x_test, y_test = data['x_test'], data['y_test']
    x_train = x_train.reshape(-1, 28*28) / 255.0
    x_test = x_test.reshape(-1, 28*28) / 255.0
    return x_train, y_train, x_test, y_test

# One-hot encoding
def one_hot(y, num_classes=10):
    return np.eye(num_classes)[y]

# Leaky ReLU
def leaky_relu(x, alpha=0.01):
    return np.where(x > 0, x, alpha * x)

def leaky_relu_derivative(x, alpha=0.01):
    return np.where(x > 0, 1, alpha)

# Softmax
def softmax(x):
    exps = np.exp(x - np.max(x, axis=1, keepdims=True))
    return exps / np.sum(exps, axis=1, keepdims=True)

# Cross-entropy loss
def cross_entropy(predictions, targets):
    epsilon = 1e-12
    predictions = np.clip(predictions, epsilon, 1. - epsilon)
    return -np.mean(np.sum(targets * np.log(predictions), axis=1))

# Accuracy
def accuracy(predictions, labels):
    return np.mean(np.argmax(predictions, axis=1) == labels)

# Initialize parameters
def init_params(input_size, hidden_size, output_size):
    params = {
        "W1": np.random.randn(input_size, hidden_size) * np.sqrt(2. / input_size),
        "b1": np.zeros((1, hidden_size)),
        "W2": np.random.randn(hidden_size, output_size) * np.sqrt(2. / hidden_size),
        "b2": np.zeros((1, output_size)),
    }
    return params

# Training loop
def train(x, y, params, epochs=10, lr=0.01, batch_size=64):
    for epoch in range(epochs):
        indices = np.arange(x.shape[0])
        np.random.shuffle(indices)
        x, y = x[indices], y[indices]

        for i in range(0, x.shape[0], batch_size):
            x_batch = x[i:i+batch_size]
            y_batch = y[i:i+batch_size]

            # Forward
            z1 = np.dot(x_batch, params["W1"]) + params["b1"]
            a1 = leaky_relu(z1)
            z2 = np.dot(a1, params["W2"]) + params["b2"]
            a2 = softmax(z2)

            # Loss
            loss = cross_entropy(a2, y_batch)

            # Backward
            dz2 = a2 - y_batch
            dW2 = np.dot(a1.T, dz2)
            db2 = np.sum(dz2, axis=0, keepdims=True)

            da1 = np.dot(dz2, params["W2"].T)
            dz1 = da1 * leaky_relu_derivative(z1)
            dW1 = np.dot(x_batch.T, dz1)
            db1 = np.sum(dz1, axis=0, keepdims=True)

            # Update
            params["W1"] -= lr * dW1 / batch_size
            params["b1"] -= lr * db1 / batch_size
            params["W2"] -= lr * dW2 / batch_size
            params["b2"] -= lr * db2 / batch_size

        print(f"Epoch {epoch+1}/{epochs}, Loss: {loss:.4f}")

    return params

# Run everything
if __name__ == "__main__":
    x_train, y_train, x_test, y_test = load_mnist()
    y_train_oh = one_hot(y_train)
    y_test_oh = one_hot(y_test)

    params = init_params(input_size=784, hidden_size=128, output_size=10)
    trained_params = train(x_train, y_train_oh, params, epochs=10, lr=0.1)

    # Evaluate
    z1 = np.dot(x_test, trained_params["W1"]) + trained_params["b1"]
    a1 = leaky_relu(z1)
    z2 = np.dot(a1, trained_params["W2"]) + trained_params["b2"]
    a2 = softmax(z2)

    acc = accuracy(a2, y_test)
    print(f"Test Accuracy: {acc:.4f}")
